#ifndef SUDOKU_BOARD_HPP
#define SUDOKU_BOARD_HPP

#include <cassert>
#include <exception>
#include <iterator>
#include <valarray>

namespace sudoku {

struct Bad_dimensions : std::exception {};

struct Board {
	typedef signed char				value_type;

	template <typename In>
	Board(In first, In last)
	    throw(Bad_dimensions);

	Board(const Board &b);

	Board &operator=(const Board &b)
	{
		Board copy = b;
		swap(copy);
		return *this;
	}

	void swap(Board &b);

	std::slice row(size_t r) const
	{
		return std::slice(r, _major, _major);
	}

	std::slice col(size_t c) const
	{
		return std::slice(c * _major, _major, 1);
	}

	// input is a square index
	std::gslice square(size_t sqi) const
	{
		size_t sz[] = { _minor, _minor };
		size_t st[] = { 1, _major };
		return std::gslice(
		    square_start(sqi),
		    std::valarray<size_t>(sz, 2),
		    std::valarray<size_t>(st, 2));
	}

	// any poss* functions slice into the 'possible' valarray

	// input is absolute
	std::slice poss(size_t i) const
	{
		return std::slice(i, _major, board.size());
	}

	std::gslice poss_row(size_t r) const
	{
		size_t size[] = { _major, _major };
		size_t stride[] = { _major, board.size() };
		return std::gslice(r,
		    std::valarray<size_t>(size, 2),
		    std::valarray<size_t>(stride, 2));
	}

	std::gslice poss_col(size_t c) const
	{
		size_t size[] = { _major, _major };
		size_t stride[] = { 1, board.size() };
		return std::gslice(c * _major,
		    std::valarray<size_t>(size, 2),
		    std::valarray<size_t>(stride, 2));
	}

	// input is square index
	std::gslice poss_square(size_t sqi) const
	{
		size_t size[] = { _minor, _minor, _major };
		size_t stride[] = { 1, _major, board.size() };
		return std::gslice(
		    square_start(sqi),
		    std::valarray<size_t>(size, 3),
		    std::valarray<size_t>(stride, 3));
	}

	// index is a square index, output is absolute
	size_t square_start(size_t sqi) const
	{
		// row is: sqi % _minor * _minor
		// col is: sqi / _minor * _minor
		// take row + col * _major
		register size_t sz =
		    (sqi % _minor + sqi / _minor * _major) * _minor;
		assert( sz < board.size() );
		return sz;
	}

	// index is absolute, output is row
	size_t which_row(size_t i) const
	{
		return i % _major;
	}

	// index is absolute, output is col
	size_t which_col(size_t i) const
	{
		return i / _major;
	}

	// index is absolute, output is square index
	size_t which_square(size_t i) const
	{
		register size_t sz =
		    which_row(i) / _minor + which_col(i) / _minor * _minor;
		assert( sz < _major );
		return sz;
	}
private:

	void init_poss();

public:
	std::valarray<value_type> board;
	std::valarray<bool> possibilities;
	size_t _major;
	size_t _minor;
};

inline void swap(Board &a, Board &b)
{
	a.swap(b);
}

template <typename In>
Board::Board(In first, In last)
    throw(Bad_dimensions) :
    board(),
    possibilities()
{
	size_t size = std::distance(first, last);
	_minor = std::pow(size, .25);
	_major = _minor * _minor;
	if (!_minor || _major * _major != size)
		throw Bad_dimensions();
	board.resize(size);
	possibilities.resize(size * _major);
	std::copy(first, last, &board[0]);
	init_poss();
}

}

#endif
