#include <cassert>

#include "board.hpp"
#include "solve.hpp"

namespace sudoku {

namespace {

// the number of possible values
inline size_t
possible_values(const Board &board, size_t i)
{
	std::slice slice = board.poss(i);
	size_t count = 0;
	for (size_t i = 0; i < slice.size(); i++) {
		size_t j = slice.start() + i * slice.stride();
		count += board.possibilities[j];
	}
	return count;
}

// returns the position of the box with the fewest possible values, or -1 if
// complete
ssize_t
next_index(const Board &board)
{
	ssize_t best = -1;
	size_t possible = board._major + 1; // one past maximum

	ssize_t sz = static_cast<ssize_t>(board.board.size());
	for (ssize_t i = 0; i < sz; i++)
		if (board.board[i] < 0) {
			size_t p = possible_values(board, i);
			if (p == 1)
				// can't get any better
				return i;
			else if (p < possible) {
				best = i;
				possible = p;
			}
		}
	return best;
}

inline void
apply_mask_1(std::valarray<bool> *possibilities, const std::gslice &slice,
    const std::valarray<bool> &mask)
{
	for (size_t i = 0; i < slice.size()[0]; i++) {
		size_t j = slice.start() + i * slice.stride()[0];

		assert(j + (slice.size()[1] - 1) * slice.stride()[1] <
		    possibilities->size());

		std::slice poss(j, slice.size()[1], slice.stride()[1]);
		(*possibilities)[poss] &= mask;
	}
}

inline void
apply_mask_2(std::valarray<bool> *possibilities, const std::gslice &slice,
    const std::valarray<bool> &mask)
{
	for (size_t r = 0; r < slice.size()[0]; r++)
		for (size_t c = 0; c < slice.size()[1]; c++) {
			size_t i = slice.start() +
			    r * slice.stride()[0] +
			    c * slice.stride()[1];

			assert(i + (slice.size()[2] - 1) * slice.stride()[2] <
			    possibilities->size());

			std::slice poss(i, slice.size()[2], slice.stride()[2]);
			(*possibilities)[poss] &= mask;
		}
}

void
apply_mask(std::valarray<bool> *possibilities, const std::gslice &slice,
    const std::valarray<bool> &mask)
{
	size_t dimensions = slice.size().size() - 1;
	switch(dimensions) {
	case 1:
		apply_mask_1(possibilities, slice, mask);
		break;
	case 2:
		apply_mask_2(possibilities, slice, mask);
		break;
	default:
		assert(0);
	}
}

void
solve_impl(std::list<boost::shared_ptr<Board> > *solutions, Board &board)
{
	ssize_t index = next_index(board);
	if (index == -1) {
		// this is a solution; board is never invalid
		solutions->push_back(
		    boost::shared_ptr<Board>(new Board(board)));
		return;
	}

	// ! 'board pass'
	std::slice poss_slice = board.poss(index);
	std::gslice poss_row_slice = board.poss_row(board.which_row(index));
	std::gslice poss_col_slice = board.poss_col(board.which_col(index));
	std::gslice poss_sq_slice = board.poss_square(
	    board.which_square(index));

	std::valarray<bool> mask(true, board._major);

	for (Board::value_type i = 0;
	    i < static_cast<Board::value_type>(poss_slice.size()); i++) {
		size_t j = poss_slice.start() + poss_slice.stride() * i;
		if (!board.possibilities[j]) continue;

		// make a copy, set the current position to the next possible
		// value 'i', and unset 'i' in the possibilities matrix
		Board copy = board;
		copy.board[index] = i;

		mask[i] = false;
		apply_mask(&copy.possibilities, poss_row_slice, mask);
		apply_mask(&copy.possibilities, poss_col_slice, mask);
		apply_mask(&copy.possibilities, poss_sq_slice, mask);
		mask[i] = true;

		solve_impl(solutions, copy);
	}
}

}

void
solve(std::list<boost::shared_ptr<Board> > *solutions, Board initial_board)
{
	solve_impl(solutions, initial_board);
}

}
